import datetime
import time
import unittest
import dateutil
import factories
from .base import BaseGrapheneElasticTestCase

__all__ = (
    'PaginationTestCase',
)


class PaginationTestCase(BaseGrapheneElasticTestCase):

    def setUp(self):
        super(PaginationTestCase, self).setUp()

        # Important thing to know about the factories.
        # The `PostFactory` factory has `num_views` between 0 and 1_000.
        # The `ManyViewsPostFactory` factory has `num_views` between
        # 2_000 and 10_000.
        self.num_elastic_posts = 22
        self.elastic_posts = factories.PostFactory.create_batch(
            self.num_elastic_posts,
            category='Elastic',
            tags=None
        )
        for _post in self.elastic_posts:
            _post.save()

        self.num_django_posts = 11
        self.django_posts = factories.PostFactory.create_batch(
            self.num_django_posts,
            category='Django'
        )
        for _post in self.django_posts:
            _post.save()

        self.num_all_posts = self.num_elastic_posts + self.num_django_posts
        self.all_posts = (
            self.elastic_posts + self.django_posts
        )

        self.num_future_users = 5
        self.future_users = factories.UserFactory.create_batch(
            self.num_future_users,
            created_at=self.faker.future_datetime()
        )
        for _post in self.future_users:
            _post.save()

        self.num_past_users = 59
        self.past_users = factories.UserFactory.create_batch(
            self.num_past_users,
            created_at=self.faker.past_datetime()
        )
        for _post in self.past_users:
            _post.save()

        self.num_all_users = self.num_past_users + self.num_future_users
        self.all_users = self.past_users + self.future_users

        time.sleep(3)

    def _test_pagination(self,
                         expected_num_results,
                         first=None,
                         last=None,
                         after=None,
                         before=None):
        """Test pagination.

        :param expected_num_results:
        :param first:
        :param last:
        :param after:
        :param before:
        :return:
        """
        _query_args_list = []
        _query_args = ""
        if first:
            _query_args_list.append('first:{}'.format(first))
        if last:
            _query_args_list.append('last:{}'.format(last))
        if after:
            _query_args_list.append('after:"{}"'.format(after))
        if before:
            _query_args_list.append('before:"{}"'.format(before))

        if _query_args_list:
            _query_args = '({})'.format(''.join(_query_args_list))

        _query = """
        {
          allPostDocuments%s {
            pageInfo {
              startCursor
              endCursor
              hasNextPage
              hasPreviousPage
            }
            edges {
              cursor
              node {
                category
                title
                content
                numViews
              }
            }
          }
        }
        """ % _query_args
        print(_query)
        executed = self.client.execute(_query)
        fields_values_sorted = []
        # TODO: Perhaps, check firsts and lasts?
        self.assertEqual(
            len(executed['data']['allPostDocuments']['edges']),
            expected_num_results
        )

    def _test_pagination_required_first_or_last(self):
        """Test pagination.

        :param expected_num_results:
        :param first:
        :param last:
        :param after:
        :param before:
        :param ordering:
        :return:
        """
        _query = """
        {
          users(ordering:{createdAt:ASC}) {
            pageInfo {
              startCursor
              endCursor
              hasNextPage
              hasPreviousPage
            }
            edges {
              cursor
              node {
                email
                firstName
                lastName
                isActive
                createdAt
              }
            }
          }
        }
        """
        print(_query)
        executed = self.client.execute(_query)
        self.assertIn('errors', executed)
        self.assertIn('message', executed['errors'][0])
        self.assertIn('`first`', executed['errors'][0]['message'])
        self.assertIn('`last`', executed['errors'][0]['message'])

    def _test_pagination_required_correct_ordering_and_limits(
            self,
            expected_num_results,
            last=None
    ):
        """Test pagination.

        :param expected_num_results:
        :param last:
        :return:
        """
        _query = """
        {
          users(ordering:{createdAt:DESC}, last:%s) {
            pageInfo {
              startCursor
              endCursor
              hasNextPage
              hasPreviousPage
            }
            edges {
              cursor
              node {
                email
                firstName
                lastName
                isActive
                createdAt
              }
            }
          }
        }
        """ % last
        print(_query)

        executed = self.client.execute(_query)
        # fields_values_sorted = []
        self.assertEqual(
            len(executed['data']['users']['edges']),
            expected_num_results
        )
        _today = datetime.date.today()
        today = datetime.datetime(
            year=_today.year,
            month=_today.month,
            day=_today.day
        )
        for edge in executed['data']['users']['edges']:
            created_at = dateutil.parser.parse(edge['node']['createdAt'])
            self.assertGreater(created_at, today)

    def test_pagination(self):
        """"Test pagination.

        :return:
        """
        with self.subTest('Test no params given, all items shall be present'):
            self._test_pagination(
                self.num_all_posts
            )

        with self.subTest('Test first 12'):
            self._test_pagination(
                first=12,
                expected_num_results=12
            )
        with self.subTest('Test no first or last params given when required'):
            self._test_pagination_required_first_or_last()

        with self.subTest('Test correct ordering and limits'):
            self._test_pagination_required_correct_ordering_and_limits(
                expected_num_results=self.num_future_users,
                last=self.num_future_users
            )


if __name__ == '__main__':
    unittest.main()



import datetime

__title__ = "graphene_elastic.helpers"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("DictionaryProxy",)


class DictionaryProxy(object):
    """Dictionary proxy."""

    def __init__(self, obj):
        if callable(obj):
            _obj = obj()
            self.__mapping = _obj.to_dict()
            self.__obj = _obj
        else:
            self.__mapping = obj.to_dict()
            self.__obj = obj

    def __getattr__(self, item):
        if item in self.__mapping:
            val = self.__mapping.get(item, None)
            if isinstance(val, datetime.datetime):
                val = val.date()
            return val

        return getattr(self.__obj, item)

    def to_dict(self):
        """To dict.

        :return:
        """
        return self.__mapping



from .filtering import *
from .search import *
from .ordering import *

__title__ = 'graphene_elastic.filter_backends'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019 Artur Barseghyan'
__license__ = 'GPL-2.0-only OR LGPL-2.1-or-later'



from ... import constants
from ..queries import (
    Term,
    Terms,
    Range,
    Exists,
    Prefix,
    Wildcard,
    GeoDistance,
    GeoPolygon,
    GeoBoundingBox,
    Contains,
    In,
    Gt,
    Gte,
    Lt,
    Lte,
    StartsWith,
    EndsWith,
    IsNull,
    Exclude,
)

__title__ = "graphene_elastic.filter_backends.filtering.queries"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = (
    'LOOKUP_FILTER_MAPPING',
)

LOOKUP_FILTER_MAPPING = {
    constants.LOOKUP_FILTER_TERM: Term,
    constants.LOOKUP_FILTER_TERMS: Terms,
    constants.LOOKUP_FILTER_RANGE: Range,
    constants.LOOKUP_FILTER_EXISTS: Exists,
    constants.LOOKUP_FILTER_PREFIX: Prefix,
    constants.LOOKUP_FILTER_WILDCARD: Wildcard,
    # constants.LOOKUP_FILTER_REGEXP: Regexp,
    # constants.LOOKUP_FILTER_FUZZY: Fuzzy,
    # constants.LOOKUP_FILTER_TYPE: Type,
    constants.LOOKUP_FILTER_GEO_DISTANCE: GeoDistance,
    constants.LOOKUP_FILTER_GEO_POLYGON: GeoPolygon,
    constants.LOOKUP_FILTER_GEO_BOUNDING_BOX: GeoBoundingBox,
    constants.LOOKUP_QUERY_CONTAINS: Contains,
    constants.LOOKUP_QUERY_IN: In,
    constants.LOOKUP_QUERY_GT: Gt,
    constants.LOOKUP_QUERY_GTE: Gte,
    constants.LOOKUP_QUERY_LT: Lt,
    constants.LOOKUP_QUERY_LTE: Lte,
    constants.LOOKUP_QUERY_STARTSWITH: StartsWith,
    constants.LOOKUP_QUERY_ENDSWITH: EndsWith,
    constants.LOOKUP_QUERY_ISNULL: IsNull,
    constants.LOOKUP_QUERY_EXCLUDE: Exclude,
}



import logging
import unittest
from faker import Faker
from graphene.test import Client
from elasticsearch import Elasticsearch
from schema import schema
from search_index.documents.settings import (
    BLOG_POST_DOCUMENT_NAME,
    SITE_USER_DOCUMENT_NAME,
)
from search_index.documents import Post, User
import factories

__all__ = (
    'BaseGrapheneElasticTestCase',
)


LOGGER = logging.getLogger(__name__)


class BaseGrapheneElasticTestCase(unittest.TestCase):
    """Base graphene-elastic test case."""

    @classmethod
    def setUpClass(cls):
        cls.client = Client(schema)
        cls.elasticsearch = Elasticsearch()
        cls.faker = Faker()

    def setUp(self):
        super(BaseGrapheneElasticTestCase, self).setUp()
        self.remove_elasticsearch_indexes()
        self.create_elasticsearch_indexes()
        # self.populate_elasticsearch_data()

    def remove_elasticsearch_indexes(self):
        """"""
        for _index in [BLOG_POST_DOCUMENT_NAME, SITE_USER_DOCUMENT_NAME]:
            try:
                _res = self.elasticsearch.indices.delete(_index)
            except Exception as err:
                print(err)

    def create_elasticsearch_indexes(self):
        """"""
        try:
            # Create the mappings in Elasticsearch
            User.init()
        except Exception as err:
            LOGGER.error(err)

        try:
            # Create the mappings in Elasticsearch
            Post.init()
        except Exception as err:
            LOGGER.error(err)
    #
    # def populate_elasticsearch_data(self):
    #     """"""
    #     posts = factories.PostFactory.create_batch(self.num_posts)
    #
    #     for post in posts:
    #         post.save()
    #
    #     users = factories.UserFactory.create_batch(self.num_users)
    #
    #     for user in users:
    #         user.save()



import time
import unittest
import factories
from .base import BaseGrapheneElasticTestCase
from ..filter_backends.queries import Direction

__all__ = (
    'OrderingBackendElasticTestCase',
)


class OrderingBackendElasticTestCase(BaseGrapheneElasticTestCase):

    def setUp(self):
        super(OrderingBackendElasticTestCase, self).setUp()

        # Important thing to know about the factories.
        # The `PostFactory` factory has `num_views` between 0 and 1_000.
        # The `ManyViewsPostFactory` factory has `num_views` between
        # 2_000 and 10_000.
        self.num_elastic_posts = 4
        self.elastic_posts = factories.PostFactory.create_batch(
            self.num_elastic_posts,
            category='Elastic',
            tags=None
        )
        for _post in self.elastic_posts:
            _post.save()

        self.num_django_posts = 3
        self.django_posts = factories.PostFactory.create_batch(
            self.num_django_posts,
            category='Django'
        )
        for _post in self.django_posts:
            _post.save()

        self.num_python_posts = 2
        self.python_posts = factories.ManyViewsPostFactory.create_batch(
            self.num_python_posts,
            category='Python',
        )
        for _post in self.python_posts:
            _post.save()

        self.num_all_posts = (
            self.num_elastic_posts +
            self.num_django_posts +
            self.num_python_posts
        )
        self.all_posts = (
            self.elastic_posts + self.django_posts + self.python_posts
        )

        time.sleep(3)

    def _test_ordering(self, field, direction):
        """Test ordering.

        :param field:
        :param direction:
        :return:
        """
        _query = """
        {
          allPostDocuments(ordering:{%s:%s}) {
            edges {
              node {
                category
                title
                content
                numViews
                tags
              }
            }
          }
        }
        """ % (field, direction.name)
        print(_query)
        executed = self.client.execute(_query)
        fields_values_sorted = []
        for edge in executed['data']['allPostDocuments']['edges']:
            field_value = edge.get('node', {}).get(field, None)
            fields_values_sorted.append(field_value)

        sorted_values = sorted(fields_values_sorted)
        if direction == Direction.DESC:
            sorted_values = list(reversed(sorted_values))

        self.assertEqual(sorted_values, fields_values_sorted)

    def _test_default_ordering(self, field, direction):
        """Test default ordering.

        :param field:
        :param direction:
        :return:
        """
        _query = """
        {
          allPostDocuments {
            edges {
              node {
                category
                title
                content
                numViews
                tags
              }
            }
          }
        }
        """
        print(_query)
        executed = self.client.execute(_query)
        fields_values_sorted = []
        for edge in executed['data']['allPostDocuments']['edges']:
            field_value = edge.get('node', {}).get(field, None)
            fields_values_sorted.append(field_value)

        sorted_values = sorted(fields_values_sorted)
        if direction == Direction.DESC:
            sorted_values = list(reversed(sorted_values))

        self.assertEqual(sorted_values, fields_values_sorted)

    def test_ordering(self):
        """"Test ordering (on fields `title` and `num_views`).

        :return:
        """
        with self.subTest('Test ordering on field `title` ascending'):
            self._test_ordering(
                'title',
                Direction.ASC
            )

        with self.subTest('Test ordering on field `title` descending'):
            self._test_ordering(
                'title',
                Direction.DESC
            )

        with self.subTest('Test ordering on field `num_views` ascending'):
            self._test_ordering(
                'numViews',
                Direction.ASC
            )

        with self.subTest('Test ordering on field `num_views` descending'):
            self._test_ordering(
                'numViews',
                Direction.DESC
            )

    def test_default_ordering(self):
        """Test default ordering (on field `num_views)."""
        with self.subTest('Test default ordering on field `num_views`'):
            self._test_default_ordering(
                'numViews',
                Direction.DESC
            )


if __name__ == '__main__':
    unittest.main()



from graphql_relay.connection.arrayconnection import (
    get_offset_with_default,
    offset_to_cursor,
)
from graphql_relay.connection.connectiontypes import (
    Connection,
    PageInfo,
    Edge,
)

from .logging import logger

__title__ = 'graphene_elastic.arrayconnection'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'connection_from_list_slice',
)


def connection_from_list_slice(
        list_slice,
        args=None,
        connection_type=None,
        edge_type=None,
        pageinfo_type=None,
        slice_start=0,
        list_length=0,
        list_slice_length=None):
    """
    Given a slice (subset) of an array, returns a connection object for use in
    GraphQL.
    This function is similar to `connectionFromArray`, but is intended for use
    cases where you know the cardinality of the connection, consider it too
    large to materialize the entire array, and instead wish pass in a slice of
    the total result large enough to cover the range specified in `args`.
    """
    connection_type = connection_type or Connection
    edge_type = edge_type or Edge
    pageinfo_type = pageinfo_type or PageInfo

    args = args or {}

    before = args.get('before')
    after = args.get('after')
    first = args.get('first')
    last = args.get('last')

    enforce_first_or_last = args.get("enforce_first_or_last")
    max_limit = args.get("max_limit")

    if enforce_first_or_last:
        assert first or last, (
            "You must provide a `first` or `last` value to properly "
            "paginate the `{}` connection."
        ).format(connection_type)

    if max_limit:
        if first or last:
            if first:
                assert first <= max_limit, (
                    "Requesting {} records on the `{}` connection exceeds "
                    "the `first` limit of {} records."
                ).format(first, connection_type, max_limit)
                first = args["first"] = min(first, max_limit)

            if last:
                assert last <= max_limit, (
                    "Requesting {} records on the `{}` connection exceeds "
                    "the `last` limit of {} records."
                ).format(last, connection_type, max_limit)
                last = args["last"] = min(last, max_limit)
        else:
            first = max_limit

    if list_slice_length is None:
        list_slice_length = len(list_slice)
    slice_end = slice_start + list_slice_length
    before_offset = get_offset_with_default(before, list_length)
    after_offset = get_offset_with_default(after, -1)

    start_offset = max(
        slice_start - 1,
        after_offset,
        -1
    ) + 1
    end_offset = min(
        slice_end,
        before_offset,
        list_length
    )
    if isinstance(first, int):
        end_offset = min(
            end_offset,
            start_offset + first
        )
    if isinstance(last, int):
        start_offset = max(
            start_offset,
            end_offset - last
        )

    # If supplied slice is too large, trim it down before mapping over it.
    _slice_qs = list_slice[
        max(start_offset - slice_start, 0):
        list_slice_length - (slice_end - end_offset)
    ]
    logger.debug(_slice_qs.to_dict())

    _slice = _slice_qs.execute()

    edges = [
        edge_type(
            node=node,
            cursor=offset_to_cursor(start_offset + i)
        )
        for i, node in enumerate(_slice)
    ]

    first_edge_cursor = edges[0].cursor if edges else None
    last_edge_cursor = edges[-1].cursor if edges else None
    lower_bound = after_offset + 1 if after else 0
    upper_bound = before_offset if before else list_length

    return connection_type(
        edges=edges,
        page_info=pageinfo_type(
            start_cursor=first_edge_cursor,
            end_cursor=last_edge_cursor,
            has_previous_page=(
                isinstance(last, int) and start_offset > lower_bound
            ),
            has_next_page=isinstance(first, int) and end_offset < upper_bound
        )
    )



__title__ = "graphene_elastic.registry"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = (
    "Registry",
    "registry",
    "get_global_registry",
    "reset_global_registry",
)


class Registry(object):
    def __init__(self):
        self._registry = {}
        self._field_registry = {}

    def register(self, cls):
        from .types import ElasticsearchObjectType

        assert issubclass(
            cls, ElasticsearchObjectType
        ), 'Only ElasticsearchObjectType can be registered, ' \
           'received "{}"'.format(
            cls.__name__
        )
        assert (
            cls._meta.registry == self
        ), "Registry for a document have to match."
        self._registry[cls._meta.document] = cls

        # Rescan all fields
        for document, cls in self._registry.items():
            cls.rescan_fields()

    def get_type_for_document(self, document):
        return self._registry.get(document)

    def register_converted_field(self, field, converted):
        self._field_registry[field] = converted

    def get_converted_field(self, field):
        return self._field_registry.get(field)


registry = None


def get_global_registry():
    global registry
    if not registry:
        registry = Registry()
    return registry


def reset_global_registry():
    global registry
    registry = None



from enum import Enum
import graphene

__title__ = "graphene_elastic.enums"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("NoValue", "convert_list_to_enum")


class NoValue(Enum):
    """String values in enum.

    Example:

    >>> class Color(NoValue):
    >>>     RED = 'stop'
    >>>     GREEN = 'go'
    >>>     BLUE = 'too fast!'

    Graphene example:

    >>> @graphene.Enum.from_enum
    >>> class ColorOptions(NoValue):
    >>>
    >>>     RED = 'stop'
    >>>     GREEN = 'go'
    >>>     BLUE = 'too fast!'
    """

    def __repr__(self):
        return "<%s.%s>" % (self.__class__.__name__, self.name)


def convert_list_to_enum(values, enum_name="DynamicEnum"):
    """Prepare list values for creating an Enum.

    Example:

    >>> values = ['red', 'green', 'blue']
    >>> print(prepare_list_for_enum(values))
    {'RED': 'red', 'GREEN': 'green', 'BLUE': 'blue'}
    """
    return NoValue(enum_name, {el.upper(): el for el in values})



"""
Transitional compatibility module. Contains various field wrappers and
helpers for painless (testing of) Elastic 6.x to Elastic 7.x transition. This
module is not supposed to solve all transition issues for you. Better move to
Elastic 7.x as soon as possible.
"""

__title__ = 'graphene_elastic.compat'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'get_elasticsearch_version',
    'nested_sort_entry',
)


def get_elasticsearch_version(default=(2, 0, 0)):
    """Get Elasticsearch version.

    :param default: Default value. Mainly added for building the docs
        when Elasticsearch is not running.
    :type default: tuple
    :return:
    :rtype: list
    """
    try:
        from elasticsearch_dsl import __version__
        return __version__
    except ImportError:
        return default


def nested_sort_entry(path):
    """String field.
        :param path: Full path to nested container, separated by period
        :type: str
        :return: Dictionary of full nested path
        :rtype: dict
        """
    version = get_elasticsearch_version()
    if version[0] < 6 or (version[0] == 6 and version[1] < 1):
        return {'nested_path': path}
    nested_path = {}
    path_list = path.split('.')
    for _ in reversed(path_list):
        if nested_path:
            nested_path = {'path': '.'.join(path_list), 'nested': nested_path}
        else:
            nested_path = {'path': '.'.join(path_list)}
        path_list.pop()
    return {'nested': nested_path}



"""
Ordering backend.
"""
import graphene
from six import string_types

from ..base import BaseBackend
from ..queries import Direction
from ...compat import nested_sort_entry

__title__ = 'graphene_elastic.filter_backends.ordering.common'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019 Artur Barseghyan'
__license__ = 'GPL 2.0/LGPL 2.1'
__all__ = (
    'DefaultOrderingFilterBackend',
    'OrderingFilterBackend',
)


class OrderingMixin(object):

    def prepare_ordering_fields(self):
        """Prepare ordering fields.

        :return: Ordering options.
        :rtype: dict
        """
        ordering_args = dict(self.args).get(self.prefix)
        if not ordering_args:
            return {}

        ordering_fields = {}

        for arg, value in ordering_args.items():
            field = self.connection_field.ordering_args_mapping.get(arg, None)
            if field is None:
                continue
            ordering_fields.update({field: {}})
            options = self.connection_field.ordering_fields.get(field)

            if options is None or isinstance(options, string_types):
                ordering_fields[field] = {
                    'field': options or field
                }
            elif 'field' not in ordering_fields[field]:
                ordering_fields[field]['field'] = field
        return ordering_fields

    @classmethod
    def transform_ordering_params(cls, ordering_params, ordering_fields):
        """Transform ordering fields to elasticsearch-dsl Search.sort()
         dictionary parameters.

        :param ordering_params: List of fields to order by.
        :param ordering_fields: Prepared ordering fields
        :type: list of str
        :type: dict
        :return: Ordering parameters.
        :rtype: list
        """
        _ordering_params = []
        if isinstance(ordering_params, dict):
            for ordering_param, ordering_direction in ordering_params.items():
                field = ordering_fields[ordering_param]
                entry = {
                    field['field']: {
                        'order': ordering_direction,
                    }
                }
                if 'path' in field:
                    entry[field['field']].update(
                        nested_sort_entry(field['path']))
                _ordering_params.append(entry)
        elif isinstance(ordering_params, (tuple, list)):
            for ordering_param in ordering_params:
                ordering_direction = Direction.ASC.value
                field = {'field': ordering_param}
                if ordering_param.startswith('-'):
                    field['field'] = ordering_param[1:]
                    ordering_direction = Direction.DESC.value
                entry = {
                    field['field']: {
                        'order': ordering_direction,
                    }
                }
                _ordering_params.append(entry)
        return _ordering_params


class OrderingFilterBackend(BaseBackend, OrderingMixin):
    """Ordering filter backend for Elasticsearch.

    Example:

        >>>   import graphene
        >>>   from graphene import Node
        >>>   from graphene_elastic import (
        >>>       ElasticsearchObjectType,
        >>>       ElasticsearchConnectionField,
        >>>   )
        >>>   from graphene_elastic.filter_backends import (
        >>>       FilteringFilterBackend,
        >>>       SearchFilterBackend,
        >>>       OrderingFilterBackend,
        >>>       DefaultOrderingFilterBackend,
        >>>   )
        >>>   from graphene_elastic.constants import (
        >>>       LOOKUP_FILTER_PREFIX,
        >>>       LOOKUP_FILTER_TERM,
        >>>       LOOKUP_FILTER_TERMS,
        >>>       LOOKUP_FILTER_WILDCARD,
        >>>       LOOKUP_QUERY_EXCLUDE,
        >>>       LOOKUP_QUERY_IN,
        >>>   )
        >>>
        >>>   from search_index.documents import Post as PostDocument
        >>>
        >>>   class Post(ElasticsearchObjectType):
        >>>
        >>>       class Meta(object):
        >>>           document = PostDocument
        >>>           interfaces = (Node,)
        >>>           filter_backends = [
        >>>               FilteringFilterBackend,
        >>>               SearchFilterBackend,
        >>>               OrderingFilterBackend,
        >>>               DefaultOrderingFilterBackend
        >>>           ]
        >>>           filter_fields = {
        >>>               'id': '_id',
        >>>               'title': {
        >>>                   'field': 'title.raw',
        >>>                   'lookups': [
        >>>                       LOOKUP_FILTER_TERM,
        >>>                       LOOKUP_FILTER_TERMS,
        >>>                       LOOKUP_FILTER_PREFIX,
        >>>                       LOOKUP_FILTER_WILDCARD,
        >>>                       LOOKUP_QUERY_IN,
        >>>                       LOOKUP_QUERY_EXCLUDE,
        >>>                   ],
        >>>                   'default_lookup': LOOKUP_FILTER_TERM,
        >>>               },
        >>>               'category': 'category.raw',
        >>>               'tags': 'tags.raw',
        >>>               'num_views': 'num_views',
        >>>           }
        >>>           search_fields = {
        >>>               'title': {'boost': 4},
        >>>               'content': {'boost': 2},
        >>>               'category': None,
        >>>           }
        >>>           ordering_fields = {
        >>>               'id': None,
        >>>               'title': 'title.raw',
        >>>               'created_at': 'created_at',
        >>>               'num_views': 'num_views',
        >>>           }
        >>>
        >>>           ordering_defaults = ('id', 'title',)

    The basic usage would be:

        query {
          allPostDocuments(ordering:{title:ASC}) {
            edges {
              node {
                title
                content
                category
                numViews
                createdAt
              }
            }
          }
        }
    """

    prefix = 'ordering'
    has_fields = True

    def field_belongs_to(self, field_name):
        return field_name in self.connection_field.ordering_fields

    def get_field_type(self, field_name, field_value, base_field_type):
        """Get field type.

        :return:
        """
        return graphene.Argument(Direction)

    def get_ordering_query_params(self):
        """Get ordering query params.

        :return: Ordering params to be used for ordering.
        :rtype: list
        """
        # TODO: Support `mode` argument.
        query_params = dict(self.args).get(self.prefix)

        if not query_params:
            return []

        ordering_query_params = dict(query_params)
        # ordering_fields is always dict
        ordering_fields = self.prepare_ordering_fields()

        return self.transform_ordering_params(
            ordering_query_params,
            ordering_fields
        )

    def filter(self, queryset):
        """Filter the queryset.

        :param queryset: Base queryset.
        :type queryset: elasticsearch_dsl.search.Search
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        ordering_query_params = self.get_ordering_query_params()
        if ordering_query_params:
            return queryset.sort(*ordering_query_params)

        return queryset


class DefaultOrderingFilterBackend(BaseBackend, OrderingMixin):
    """Default ordering filter backend for Elasticsearch.

    Make sure this is your last ordering backend.

    Example:

        >>>   import graphene
        >>>   from graphene import Node
        >>>   from graphene_elastic import (
        >>>       ElasticsearchObjectType,
        >>>       ElasticsearchConnectionField,
        >>>   )
        >>>   from graphene_elastic.filter_backends import (
        >>>       FilteringFilterBackend,
        >>>       SearchFilterBackend,
        >>>       OrderingFilterBackend,
        >>>       DefaultOrderingFilterBackend,
        >>>   )
        >>>   from graphene_elastic.constants import (
        >>>       LOOKUP_FILTER_PREFIX,
        >>>       LOOKUP_FILTER_TERM,
        >>>       LOOKUP_FILTER_TERMS,
        >>>       LOOKUP_FILTER_WILDCARD,
        >>>       LOOKUP_QUERY_EXCLUDE,
        >>>       LOOKUP_QUERY_IN,
        >>>   )
        >>>
        >>>   from search_index.documents import Post as PostDocument
        >>>
        >>>   class Post(ElasticsearchObjectType):
        >>>
        >>>       class Meta(object):
        >>>           document = PostDocument
        >>>           interfaces = (Node,)
        >>>           filter_backends = [
        >>>               FilteringFilterBackend,
        >>>               SearchFilterBackend,
        >>>               OrderingFilterBackend,
        >>>               DefaultOrderingFilterBackend
        >>>           ]
        >>>           filter_fields = {
        >>>               'id': '_id',
        >>>               'title': {
        >>>                   'field': 'title.raw',
        >>>                   'lookups': [
        >>>                       LOOKUP_FILTER_TERM,
        >>>                       LOOKUP_FILTER_TERMS,
        >>>                       LOOKUP_FILTER_PREFIX,
        >>>                       LOOKUP_FILTER_WILDCARD,
        >>>                       LOOKUP_QUERY_IN,
        >>>                       LOOKUP_QUERY_EXCLUDE,
        >>>                   ],
        >>>                   'default_lookup': LOOKUP_FILTER_TERM,
        >>>               },
        >>>               'category': 'category.raw',
        >>>               'tags': 'tags.raw',
        >>>               'num_views': 'num_views',
        >>>           }
        >>>           search_fields = {
        >>>               'title': {'boost': 4},
        >>>               'content': {'boost': 2},
        >>>               'category': None,
        >>>           }
        >>>           ordering_fields = {
        >>>               'id': None,
        >>>               'title': 'title.raw',
        >>>               'created_at': 'created_at',
        >>>               'num_views': 'num_views',
        >>>           }
        >>>
        >>>           ordering_defaults = ('id', 'title',)
    """

    prefix = 'ordering'
    has_fields = False

    def get_ordering_query_params(self):
        """Get ordering query params.

        :return: Ordering params to be used for ordering.
        :rtype: list
        """
        query_params = dict(self.args).get(self.prefix)
        ordering_query_params = dict(query_params) if query_params else {}
        ordering_params_present = False
        # Remove invalid ordering query params
        for query_param, ordering_direction in ordering_query_params.items():
            if query_param in self.connection_field.ordering_fields:
                ordering_params_present = True
                break

        # If no valid ordering params specified, fall back to `view.ordering`
        if not ordering_params_present:
            return self.get_default_ordering_params()

        return {}

    def get_default_ordering_params(self):
        """Get the default ordering params for the view.

        :return: Ordering params to be used for ordering.
        :rtype: list
        """
        ordering = getattr(self.connection_field, 'ordering_defaults', None)

        if isinstance(ordering, string_types):
            ordering = [ordering]
        # For backwards compatibility require
        # default ordering to be keys in ordering_fields not field value
        # in order to be properly transformed
        if (
            ordering is not None
            and hasattr(self.connection_field, 'ordering_fields')
            and self.connection_field.ordering_fields is not None
            and all(
                field.lstrip('-') in self.connection_field.ordering_fields
                for field in ordering
            )
        ):
            return self.transform_ordering_params(
                ordering,
                self.prepare_ordering_fields()
            )
        return ordering

    def filter(self, queryset):
        """Filter the queryset.

        :param queryset: Base queryset.
        :type queryset: elasticsearch_dsl.search.Search
        :return: Updated queryset.
        :rtype: elasticsearch_dsl.search.Search
        """
        ordering_query_params = self.get_ordering_query_params()

        if ordering_query_params:
            return queryset.sort(*ordering_query_params)

        return queryset



import datetime
import json

from elasticsearch_dsl import InnerDoc, Document, AttrDict, AttrList
from graphene.types.json import JSONString as OriginalJSONString

__title__ = 'graphene_elastic.types.json_string'
__author__ = 'Artur Barseghyan <artur.barseghyan@gmail.com>'
__copyright__ = '2019 Artur Barseghyan'
__license__ = 'GPL-2.0-only OR LGPL-2.1-or-later'
__all__ = (
    'JSONString',
    'to_serializable',
)


def to_serializable(o):
    if isinstance(o, (datetime.date, datetime.datetime)):
        return o.isoformat()
    elif isinstance(o, (InnerDoc, Document)):
        return o.to_dict()
    elif isinstance(o, AttrList):
        return [to_serializable(_l) for _l in o._l_]
    elif isinstance(o, AttrDict):
        return o.to_dict()
    elif isinstance(o, list):
        return [to_serializable(_l) for _l in o]

    return o


class JSONString(OriginalJSONString):
    @staticmethod
    def serialize(dt):
        # return to_serializable(dt)
        return json.loads(json.dumps(dt, default=to_serializable))



import time
import unittest
from graphene.utils.str_converters import to_camel_case
import factories
from ..constants import (
    LOOKUP_FILTER_TERM,
    LOOKUP_FILTER_TERMS,
    LOOKUP_FILTER_RANGE,
    LOOKUP_FILTER_EXISTS,
    LOOKUP_FILTER_PREFIX,
    LOOKUP_FILTER_WILDCARD,
    # LOOKUP_FILTER_REGEXP,
    # LOOKUP_FILTER_FUZZY,
    # LOOKUP_FILTER_TYPE,
    LOOKUP_FILTER_GEO_DISTANCE,
    LOOKUP_FILTER_GEO_POLYGON,
    LOOKUP_FILTER_GEO_BOUNDING_BOX,
    LOOKUP_QUERY_CONTAINS,
    LOOKUP_FILTER_PREFIX,
    LOOKUP_QUERY_IN,
    LOOKUP_QUERY_GT,
    LOOKUP_QUERY_GTE,
    LOOKUP_QUERY_LT,
    LOOKUP_QUERY_LTE,
    LOOKUP_QUERY_STARTSWITH,
    LOOKUP_QUERY_ENDSWITH,
    LOOKUP_QUERY_ISNULL,
    LOOKUP_QUERY_EXCLUDE,
    VALUE,
    LOWER,
    UPPER,
)
from .base import BaseGrapheneElasticTestCase

__all__ = (
    'FilterBackendElasticTestCase',
)


class FilterBackendElasticTestCase(BaseGrapheneElasticTestCase):

    def setUp(self):
        super(FilterBackendElasticTestCase, self).setUp()

        # Important thing to know about the factories.
        # The `PostFactory` factory has `num_views` between 0 and 1_000.
        # The `ManyViewsPostFactory` factory has `num_views` between
        # 2_000 and 10_000.
        self.num_elastic_posts = 4
        self.elastic_posts = factories.PostFactory.create_batch(
            self.num_elastic_posts,
            category='Elastic',
            tags=None
        )
        for _post in self.elastic_posts:
            _post.save()

        self.num_django_posts = 3
        self.django_posts = factories.PostFactory.create_batch(
            self.num_django_posts,
            category='Django'
        )
        for _post in self.django_posts:
            _post.save()

        self.num_python_posts = 2
        self.python_posts = factories.ManyViewsPostFactory.create_batch(
            self.num_python_posts,
            category='Python',
        )
        for _post in self.python_posts:
            _post.save()

        self.num_all_posts = (
            self.num_elastic_posts +
            self.num_django_posts +
            self.num_python_posts
        )
        self.all_posts = (
            self.elastic_posts + self.django_posts + self.python_posts
        )

        time.sleep(3)

    def _test_filter_text_lookups(self,
                                  query,
                                  num_posts,
                                  lookup=VALUE,
                                  field='category'):
        """Test filter text lookups (on field `category`).

        :param query:
        :param num_posts:
        :param field:
        :return:
        """
        _query = """
        query {
          allPostDocuments(filter:{%s:{%s:%s}}) {
            edges {
              node {
                category
                title
                content
                numViews
                comments
              }
            }
          }
        }
        """ % (field, lookup, query)
        print(_query)
        executed = self.client.execute(_query)
        self.assertEqual(
            len(executed['data']['allPostDocuments']['edges']),
            num_posts
        )

    def _test_filter_number_lookups(self,
                                    value,
                                    num_posts,
                                    lookup=LOOKUP_QUERY_GT):
        """Test filter number lookups (on field `num_views`).

        :param value:
        :param num_posts:
        :param lookup:
        :return:
        """
        _query = """
        query {
          allPostDocuments(filter:{numViews:{%s:%s}}) {
            edges {
              node {
                category
                title
                content
                numViews
                comments
              }
            }
          }
        }
        """ % (lookup, value)
        executed = self.client.execute(_query)
        self.assertEqual(
            len(executed['data']['allPostDocuments']['edges']),
            num_posts
        )

    def test_filter_term_terms_lookup(self):
        """"Test filter `term` and `terms` lookups (on field `category`).

        :return:
        """
        with self.subTest('Test filter on field `category` "Django" '
                          'using default lookup (`term`)'):
            self._test_filter_text_lookups(
                '"Django"',
                self.num_django_posts
            )

        with self.subTest('Test filter on field `category` "Django" '
                          'using `term` lookup'):
            self._test_filter_text_lookups(
                '"Django"',
                self.num_django_posts,
                lookup=LOOKUP_FILTER_TERM
            )

        with self.subTest('Test filter on field `category` "Elastic" '
                          'using default lookup (`term`)'):
            self._test_filter_text_lookups(
                '"Elastic"',
                self.num_elastic_posts
            )

        with self.subTest('Test filter on field `category` "Elastic" '
                          'using `term` lookup'):
            self._test_filter_text_lookups(
                '"Elastic"',
                self.num_elastic_posts,
                LOOKUP_FILTER_TERM
            )

        with self.subTest('Test filter on field `category` '
                          '["Elastic", "Django"] using `terms` lookup'):
            self._test_filter_text_lookups(
                '["Elastic", "Django"]',
                self.num_elastic_posts + self.num_django_posts,
                LOOKUP_FILTER_TERMS
            )

        with self.subTest('Test filter on field `category` '
                          '["Elastic", "Django"] using `in` lookup'):
            self._test_filter_text_lookups(
                '["Elastic", "Django"]',
                self.num_elastic_posts + self.num_django_posts,
                LOOKUP_QUERY_IN
            )

    def test_filter_prefix_starts_ends_with_contains_wildcard_lookups(self):
        """"Test filters `prefix`, `starts_with` and `ends_with` lookups (on
        field `category`).

        :return:
        """
        with self.subTest('Test filter on field `category` "Elastic" '
                          'using `contains` lookup'):
            self._test_filter_text_lookups(
                '"ytho"',
                self.num_python_posts,
                LOOKUP_QUERY_CONTAINS
            )

        with self.subTest('Test filter on field `category` "Elastic" '
                          'using `wildcard` lookup'):
            self._test_filter_text_lookups(
                '"*ytho*"',
                self.num_python_posts,
                LOOKUP_FILTER_WILDCARD
            )

        with self.subTest('Test filter on field `category` "Elastic" '
                          'using `prefix` lookup'):
            self._test_filter_text_lookups(
                '"Pyth"',
                self.num_python_posts,
                to_camel_case(LOOKUP_FILTER_PREFIX)
            )

        with self.subTest('Test filter on field `category` "Elastic" '
                          'using `starts_with` lookup'):
            self._test_filter_text_lookups(
                '"Pyth"',
                self.num_python_posts,
                to_camel_case(LOOKUP_QUERY_STARTSWITH)
            )

        with self.subTest('Test filter on field `category` "Elastic" '
                          'using `ends_with` lookup'):
            self._test_filter_text_lookups(
                '"ython"',
                self.num_python_posts,
                to_camel_case(LOOKUP_QUERY_ENDSWITH)
            )

    def test_filter_exclude_lookup(self):
        """"Test filter `exclude` lookup (on field `category`).

        :return:
        """
        with self.subTest('Test filter on field `category` "Elastic" '
                          'using `exclude` lookup'):
            self._test_filter_text_lookups(
                '"Python"',
                self.num_all_posts - self.num_python_posts,
                LOOKUP_QUERY_EXCLUDE
            )

    def test_filter_exists_is_null_lookups(self):
        """"Test filter `exists` lookup (on fields `category`
        and `i_do_not_exist`).

        :return:
        """
        with self.subTest('Test filter on field `category`'
                          'using `exists` lookup'):
            self._test_filter_text_lookups(
                'true',
                self.num_all_posts,
                LOOKUP_FILTER_EXISTS
            )

        with self.subTest('Test filter on field `category`'
                          'using `is_null` lookup'):
            self._test_filter_text_lookups(
                'false',
                self.num_all_posts,
                to_camel_case(LOOKUP_QUERY_ISNULL)
            )

        # TODO: See if we can test this case
        # with self.subTest('Test filter on field `i_do_not_exist`'
        #                   'using `exists` lookup'):
        #     self._test_filter_text_lookups(
        #         'true',
        #         0,
        #         LOOKUP_FILTER_EXISTS,
        #         field='i_do_not_exist'
        #     )

    def test_filter_gt_gte_lt_lte_range_lookups(self):
        """"Test filter `gt`, `gte`, `lt`, `lte`, `range` lookups (on
        field `num_views`).

        :return:
        """
        # This should be all posts, since minimum value for posts is 0.
        with self.subTest('Test filter on field `num_views` '
                          'using `gt` lookup'):
            self._test_filter_number_lookups(
                '"0.1"',
                self.num_all_posts
            )

        # This should be Python posts only, since they may start at 2_000.
        with self.subTest('Test filter on field `num_views` '
                          'using `gt` lookup'):
            self._test_filter_number_lookups(
                '"1999"',
                self.num_python_posts
            )

        # This shall be all posts (including 0).
        with self.subTest('Test filter on field `num_views` '
                          'using `gte` lookup'):
            self._test_filter_number_lookups(
                '"0"',
                self.num_all_posts,
                lookup=LOOKUP_QUERY_GTE
            )

        # This shall be Python posts only, since they may start at 2_000.
        with self.subTest('Test filter on field `num_views` '
                          'using `gte` lookup'):
            self._test_filter_number_lookups(
                '"2000"',
                self.num_python_posts,
                lookup=LOOKUP_QUERY_GTE
            )

        # This shall be all posts, since maximum is 10_000.
        with self.subTest('Test filter on field `num_views` '
                          'using `lt` lookup'):
            self._test_filter_number_lookups(
                '"10001"',
                self.num_all_posts,
                lookup=LOOKUP_QUERY_LT
            )

        # This shall exclude Python posts, since they start at 2_000.
        with self.subTest('Test filter on field `num_views` '
                          'using `lt` lookup'):
            self._test_filter_number_lookups(
                '"2000"',
                self.num_all_posts - self.num_python_posts,
                lookup=LOOKUP_QUERY_LT
            )

        # This shall be all posts, since maximum is 10_000.
        with self.subTest('Test filter on field `num_views` '
                          'using `lte` lookup'):
            self._test_filter_number_lookups(
                '"10000"',
                self.num_all_posts,
                lookup=LOOKUP_QUERY_LTE
            )

        # This shall exclude all Python posts, since they start at 2_000
        with self.subTest('Test filter on field `num_views` '
                          'using `lte` lookup'):
            self._test_filter_number_lookups(
                '"1999"',
                self.num_all_posts - self.num_python_posts,
                lookup=LOOKUP_QUERY_LTE
            )

        # To test range successfully, since we do not make specific range
        # factories in between, we simply count the number of posts
        # between 100 and 300 and test.
        with self.subTest('Test filter on field `num_views` '
                          'using `range` lookup'):
            _count = 0
            for _p in self.all_posts:
                if 100 <= _p.num_views <= 300:
                    _count += 1
            self._test_filter_number_lookups(
                '{%s: "%s", %s: "%s"}' % (LOWER, '100', UPPER, '300'),
                _count,
                lookup=LOOKUP_FILTER_RANGE
            )


if __name__ == '__main__':
    unittest.main()



import base64
import graphene


class FileFieldType(graphene.ObjectType):

    content_type = graphene.String()
    md5 = graphene.String()
    chunk_size = graphene.Int()
    length = graphene.Int()
    data = graphene.String()

    @classmethod
    def _resolve_fs_field(cls, field, name, default_value=None):
        v = getattr(field.instance, field.key)
        return getattr(v, name, default_value)

    def resolve_content_type(self, info):
        return FileFieldType._resolve_fs_field(self, "content_type")

    def resolve_md5(self, info):
        return FileFieldType._resolve_fs_field(self, "md5")

    def resolve_chunk_size(self, info):
        return FileFieldType._resolve_fs_field(self, "chunk_size", 0)

    def resolve_length(self, info):
        return FileFieldType._resolve_fs_field(self, "length", 0)

    def resolve_data(self, info):
        v = getattr(self.instance, self.key)
        data = v.read()
        if data is not None:
            return base64.b64encode(data)
        return None


class _CoordinatesTypeField(graphene.ObjectType):

    type = graphene.String()

    def resolve_type(self, info):
        return self["type"]

    def resolve_coordinates(self, info):
        return self["coordinates"]


class PointFieldType(_CoordinatesTypeField):

    coordinates = graphene.List(graphene.Float)


class PolygonFieldType(_CoordinatesTypeField):

    coordinates = graphene.List(graphene.List(graphene.List(graphene.Float)))


class MultiPolygonFieldType(_CoordinatesTypeField):

    coordinates = graphene.List(
        graphene.List(graphene.List(graphene.List(graphene.Float)))
    )



import graphene

from ..constants import (
    DYNAMIC_CLASS_NAME_PREFIX,
    SEPARATOR_LOOKUP_NAME,
    SEPARATOR_LOOKUP_FILTER,
    SEPARATOR_LOOKUP_COMPLEX_VALUE,
    SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE,
)

__title__ = "graphene_elastic.filter_backends.base"
__author__ = "Artur Barseghyan <artur.barseghyan@gmail.com>"
__copyright__ = "2019 Artur Barseghyan"
__license__ = "GPL-2.0-only OR LGPL-2.1-or-later"
__all__ = ("BaseBackend",)


class BaseBackend(object):

    prefix = None
    has_fields = False

    def __init__(self, connection_field, args=None):
        self.connection_field = connection_field
        self.args = args or {}
        assert self.prefix

    def field_belongs_to(self, field_name):
        raise NotImplementedError

    def add_arg_prefix(self, arg_name):
        return "{}_{}".format(self.prefix, arg_name)

    def get_field_name(self, arg_name):
        if arg_name.startswith(self.prefix):
            return arg_name.lstrip("{}_".format(self.prefix))

    def arg_belongs_to(self, arg_name):
        field_name = self.get_field_name(arg_name)
        if field_name:
            return self.field_belongs_to(field_name)
        return False

    def filter(self, queryset):
        raise NotImplementedError

    @property
    def filter_fields(self):
        return self.connection_field.type._meta.node._meta.filter_fields

    @property
    def search_fields(self):
        return self.connection_field.type._meta.node._meta.search_fields

    @property
    def doc_type(self):
        return self.connection_field.document._doc_type

    @classmethod
    def generic_fields(cls):
        """Generic backend specific fields.

        For instance, for search filter backend it would be
        ``{'search': String()}``.

        :return:
        :rtype dict:
        """
        return {}

    def get_field_type(self, field_name, field_value, base_field_type):
        """Get field type.

        :return:
        """
        raise NotImplementedError

    def get_backend_default_fields_params(self):
        """Backend default filter params.

        :rtype: dict
        :return:
        """
        return {}

    def get_backend_fields(self, items, is_filterable_func, get_type_func):
        """Construct backend fields.

        :param items:
        :param is_filterable_func:
        :param get_type_func:
        :return:
        """
        params = self.get_backend_default_fields_params()
        for _k, _v in items:
            if is_filterable_func(_k):
                # Getting other backend specific fields (schema dependant)
                if self.field_belongs_to(_k):
                    params.update(
                        {_k: self.get_field_type(_k, _v, get_type_func(_v))}
                    )

        return {
            self.prefix: graphene.Argument(
                type(
                    "{}{}{}BackendFilter".format(
                        DYNAMIC_CLASS_NAME_PREFIX,
                        self.prefix.title(),
                        self.connection_field.type.__name__
                    ),
                    (graphene.InputObjectType,),
                    params,
                )
            )
        }

    @classmethod
    def split_lookup_name(cls, value, maxsplit=-1):
        """Split lookup value.

        :param value: Value to split.
        :param maxsplit: The `maxsplit` option of `string.split`.
        :type value: str
        :type maxsplit: int
        :return: Lookup value split into a list.
        :rtype: list
        """
        return value.split(SEPARATOR_LOOKUP_NAME, maxsplit)

    @classmethod
    def split_lookup_filter(cls, value, maxsplit=-1):
        """Split lookup filter.

        :param value: Value to split.
        :param maxsplit: The `maxsplit` option of `string.split`.
        :type value: str
        :type maxsplit: int
        :return: Lookup filter split into a list.
        :rtype: list
        """
        return value.split(SEPARATOR_LOOKUP_FILTER, maxsplit)

    @classmethod
    def split_lookup_complex_value(cls, value, maxsplit=-1):
        """Split lookup complex value.

        :param value: Value to split.
        :param maxsplit: The `maxsplit` option of `string.split`.
        :type value: str
        :type maxsplit: int
        :return: Lookup filter split into a list.
        :rtype: list
        """
        return value.split(SEPARATOR_LOOKUP_COMPLEX_VALUE, maxsplit)

    @classmethod
    def split_lookup_complex_multiple_value(cls, value, maxsplit=-1):
        """Split lookup complex multiple value.

        :param value: Value to split.
        :param maxsplit: The `maxsplit` option of `string.split`.
        :type value: str
        :type maxsplit: int
        :return: Lookup filter split into a list.
        :rtype: list
        """
        return value.split(SEPARATOR_LOOKUP_COMPLEX_MULTIPLE_VALUE, maxsplit)

    @classmethod
    def apply_filter(cls, queryset, options=None, args=None, kwargs=None):
        """Apply filter.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return queryset.filter(*args, **kwargs)

    @classmethod
    def apply_query(cls, queryset, options=None, args=None, kwargs=None):
        """Apply query.

        :param queryset:
        :param options:
        :param args:
        :param kwargs:
        :return:
        """
        if args is None:
            args = []
        if kwargs is None:
            kwargs = {}
        return queryset.query(*args, **kwargs)
